/*
 *	netbug.c
 *
 * - the global data structures for the modified system library
 *
 * Received (read, recv, recvfrom) packets can be 
 * + dropped entirely with probability ``dropchance'',
 * + stored for later delivery with probability ``storechance''.
 *   - stored packets will be duplicated with probability ``dupchance''.
 *   - stored packets will be delivered after arrival of the next packet on
 *     the socket with probability ``misorderchance''.
 *
 * int recvfrom(buf, from) {
 *   int ret, tmpret;
 *   char *tmpbuf;
 *   struct sockaddr_in tmpfrom;
 *
 *   if (socket_and_addr_are_appropriate_to_disturb) {
 *     if (chance() < misorderchance
 *         && there_is_a_stored_packet_for_this_socket) {
 *       ret = stored_len; buf = stored_buf; from = stored_from;
 *       if (chance() > dupchance) remove_the_stored_packet_from_buffer;
 *       return ret;
 *     }
 *     tmpret = syscall(SYS_recvfrom, tmpbuf, tmpfrom);
 *     if (chance() < storechance) store_received_packet;
 *     if (chance() > dropchance) {
 *       ret = tmpret; buf = tmpbuf; from = tmpfrom;
 *       return ret;
 *     }
 *   }
 *   return syscall(SYS_recvfrom, buf, from);
 * }
 *
 * Outgoing packets (write, send, sendto) can also be dropped, duplicated
 * or misordered. The behavioural difference in disturbing sending vs.
 * disturbing receiving gets significant mainly in case of multicasts /
 * broadcasts.
 *
 * int sendto(packet, len, addr) {
 *   int ret;
 *   int drop = false;
 *
 *   if (socket_and_addr_are_appropriate_to_disturb) {
 *     if (chance() < dropchance) drop = true;
 *     if (chance() < misorderchance
 *         && there_is_a_stored_packet_for_this_socket) {
 *       if (!drop) ret = syscall(SYS_sendto, packet, len, addr);
 *       else ret = len;
 *       syscall(SYS_sendto, stored_packet, stored_len, stored_addr);
 *       if (chance() > dupchance) remove_the_stored_packet_from_buffer;
 *     }
 *     if (drop) return ret;
 *     if (chance() < storechance) {
 *       store_packet_in_free_buffer_but_drop_if_none_free;
 *       if (chance() > dupchance) return ret;
 *     }
 *   }
 *   return syscall(SYS_sendto, packet);
 * }
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "netbug.h"

#ifdef DEBUG
int debug = 0;
#endif


/*
 *	some global variables
 */

int	_netbug_inited = 0;
int	_netbug_seed = -1;
int	_netbug_verbose = 0;

/* we can store up to 5 packets for delayed delivery in each direction */
struct _netbug_buf _netbug_recvbufs[5], _netbug_sendbufs[5];

int	_netbug_recv_delay = 0;	/* uniform distribution [0..x] */
double	_netbug_recv_dropchance = 0.0;
double	_netbug_recv_storechance = 0.0;
double	_netbug_recv_dupchance = 0.0;
double	_netbug_recv_misorderchance = 0.0;
int	_netbug_recv_maxsize = 0;		  /* no restrictions */
int	_netbug_send_delay = 0;	/* uniform distribution [0..x] */
double	_netbug_send_dropchance = 0.0;
double	_netbug_send_storechance = 0.0;
double	_netbug_send_dupchance = 0.0;
double	_netbug_send_misorderchance = 0.0;
int	_netbug_send_maxsize = 0;		  /* no restrictions */

int	_netbug_recvd = 0;
int	_netbug_recv_delayed = 0;
int	_netbug_recv_dropped = 0;
int	_netbug_recv_stored = 0;
int	_netbug_recv_duped = 0;
int	_netbug_recv_misordered = 0;
int	_netbug_recvd_maxsize = 0;
int	_netbug_sent = 0;
int	_netbug_send_delayed = 0;
int	_netbug_send_dropped = 0;
int	_netbug_send_stored = 0;
int	_netbug_send_duped = 0;
int	_netbug_send_misordered = 0;
int	_netbug_sent_maxsize = 0;



/* returns uniformly distributed random number from [0..1] */
double _netbug_chance() {
  /* random() returns a long in [0 .. 2**31-1] */
  return (double)random() / (double)(MAXLONG);
}
int _netbug_uniform(upper) int upper; {
  /* random() returns a long in [0 .. 2**31-1] */
  return (int)((double)random() / (double)(MAXLONG) * (double)upper);
}

struct _netbug_buf *_netbug_search_stored_recv_packet(s) int s; {
  int i;
  for (i = 0; i < NETBUG_NBUFS; i++)
    if (_netbug_recvbufs[i].fd == s) return _netbug_recvbufs + i;
  return NULL;
}

struct _netbug_buf *_netbug_search_free_recv_packet() {
  int i;
  for (i = 0; i < NETBUG_NBUFS; i++)
    if (_netbug_recvbufs[i].fd == -1) return _netbug_recvbufs + i;
  return NULL;
}

struct _netbug_buf *_netbug_search_stored_send_packet(s) int s; {
  int i;
  for (i = 0; i < NETBUG_NBUFS; i++)
    if (_netbug_sendbufs[i].fd == s) return _netbug_sendbufs + i;
  return NULL;
}

struct _netbug_buf *_netbug_search_free_send_packet() {
  int i;
  for (i = 0; i < NETBUG_NBUFS; i++)
    if (_netbug_sendbufs[i].fd == -1) return _netbug_sendbufs + i;
  return NULL;
}

#define TYP_INT 1
#define TYP_DOUBLE 2

static struct envvars {
  char *varname;
  int vartype;
  POINTER var;
} envvars[] = {
  { "NETBUG_VERBOSE", TYP_INT, &_netbug_verbose },
  { "NETBUG_SEED", TYP_INT, &_netbug_seed },
  { "NETBUG_RECV_DELAY", TYP_INT, &_netbug_recv_delay },
  { "NETBUG_RECV_DROPCHANCE", TYP_DOUBLE, &_netbug_recv_dropchance },
  { "NETBUG_RECV_STORECHANCE", TYP_DOUBLE, &_netbug_recv_storechance },
  { "NETBUG_RECV_MISORDERCHANCE", TYP_DOUBLE, &_netbug_recv_misorderchance },
  { "NETBUG_RECV_DUPCHANCE", TYP_DOUBLE, &_netbug_recv_dupchance },
  { "NETBUG_RECV_MAXSIZE", TYP_INT, &_netbug_recv_maxsize },
  { "NETBUG_SEND_DELAY", TYP_INT, &_netbug_send_delay },
  { "NETBUG_SEND_DROPCHANCE", TYP_DOUBLE, &_netbug_send_dropchance },
  { "NETBUG_SEND_STORECHANCE", TYP_DOUBLE, &_netbug_send_storechance },
  { "NETBUG_SEND_MISORDERCHANCE", TYP_DOUBLE, &_netbug_send_misorderchance },
  { "NETBUG_SEND_DUPCHANCE", TYP_DOUBLE, &_netbug_send_dupchance },
  { "NETBUG_SEND_MAXSIZE", TYP_INT, &_netbug_send_maxsize },
  { NULL, 0, NULL }
};

void _netbug_init() {
  struct envvars *e;
  char *s;
  int i;

  if (_netbug_inited) return;

  _netbug_inited = 1;

  for (i = 0; i < NETBUG_NBUFS; i++)
    _netbug_recvbufs[i].fd = _netbug_sendbufs[i].fd = -1;

  for (e = envvars; e->varname; e++) {
    s = getenv(e->varname);
    if (s) 
      switch(e->vartype) {
      case TYP_INT:	*(int *)e->var = atoi(s); break;
      case TYP_DOUBLE:	*(double *)e->var = atof(s); break;
      default:		abort(); break;
      }
  }
  
  if (-1 == _netbug_seed) srandom(time(NULL));
  else srandom(_netbug_seed);

  if (_netbug_verbose)
  {
      fprintf (stderr,
          "_netbug_seed = \t%d\n"
          "_netbug_recv_delay = \t%d\n"
          "_netbug_recv_dropchance = \t%f\n"
          "_netbug_recv_storechance = \t%f\n"
          "_netbug_recv_dupchance = \t%f\n"
          "_netbug_recv_misorderchance = \t%f\n"
          "_netbug_recv_maxsize = \t%d\n"
          "_netbug_send_delay = \t%d\n"
          "_netbug_send_dropchance = \t%f\n"
          "_netbug_send_storechance = \t%f\n"
          "_netbug_send_dupchance = \t%f\n"
          "_netbug_send_misorderchance = \t%f\n"
          "_netbug_send_maxsize = \t%d\n",
          _netbug_seed, 
          _netbug_recv_delay,
          _netbug_recv_dropchance,
          _netbug_recv_storechance,
          _netbug_recv_dupchance,
          _netbug_recv_misorderchance,
          _netbug_recv_maxsize,
          _netbug_send_delay,
          _netbug_send_dropchance,
          _netbug_send_storechance,
          _netbug_send_dupchance,
          _netbug_send_misorderchance,
          _netbug_send_maxsize);
  }
}

void _netbug_stats () 
{
    fprintf (stderr,
             "_netbug_recvd = \t%d\n"
             "_netbug_recv_delayed = \t%d\n"
             "_netbug_recv_dropped = \t%d\n"
             "_netbug_recv_stored = \t%d\n"
             "_netbug_recv_duped = \t%d\n"
             "_netbug_recv_misordered = \t%d\n"
             "_netbug_recvd_maxsize = \t%d\n"
             "_netbug_sent = \t%d\n"
             "_netbug_send_delayed = \t%d\n"
             "_netbug_send_dropped = \t%d\n"
             "_netbug_send_stored = \t%d\n"
             "_netbug_send_duped = \t%d\n"
             "_netbug_send_misordered = \t%d\n"
             "_netbug_sent_maxsize = \t%d\n",
             _netbug_recvd,
             _netbug_recv_delayed,
             _netbug_recv_dropped,
             _netbug_recv_stored,
             _netbug_recv_duped,
             _netbug_recv_misordered,
             _netbug_recvd_maxsize,
             _netbug_sent,
             _netbug_send_delayed,
             _netbug_send_dropped,
             _netbug_send_stored,
             _netbug_send_duped,
             _netbug_send_misordered,
             _netbug_sent_maxsize);
}



