#include "rblock.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <assert.h>

t_rblock* rblock_new(t_enBlockMarker marker) {
	t_rblock* blk = malloc(RBLOCK_SIZE_LOCAL);
	memset(blk, 0, RBLOCK_SIZE_LOCAL);
	blk->header.marker = marker;
	return blk;
}

t_rblock* rblock_new_with_space(t_enBlockMarker marker, size_t space) {
	t_rblock* blk = rblock_new(marker);

	blk->payload = malloc(space);
	memset(blk->payload, 0, space);
	blk->payload_size = space;
	return blk;
}

void rblock_free(t_rblock* blk) {
	if (blk != NULL) {
		free(blk->payload);
		free(blk);
	}
}

void rblock_copy(t_rblock* src, t_rblock* dest) {
	memcpy(dest, src, RBLOCK_SIZE_LOCAL);
}

#ifndef WORKAROUND_BUGGY_LIBNETBUG
size_t rblock_send(int fd, t_rblock* block, int flags) {
	struct iovec iov[2];
	struct msghdr msg_hdr;

	if (block == NULL) {
		return -1;
	}

	iov[0].iov_base	= block;
	iov[0].iov_len	= RBLOCK_SIZE_STREAMED;
	iov[1].iov_base	= block->payload;
	iov[1].iov_len	= block->payload_size;

	memset(&msg_hdr, 0, sizeof(struct msghdr));

	msg_hdr.msg_iov	= iov;
	if ( (block->payload==NULL) || (block->payload_size == 0)) {
		msg_hdr.msg_iovlen = 1;
	} else {
		msg_hdr.msg_iovlen = 2;
	}

	return sendmsg(fd, &msg_hdr, flags);
}

int rblock_recv_into(int fd, t_rblock* block, int flags, char payloadbuf[], size_t bufsize) {
	struct iovec iov[2];
	struct msghdr msg_hdr;
	size_t size;

	memset(&msg_hdr, 0, sizeof(struct msghdr));

	iov[0].iov_base	= block;
	iov[0].iov_len	= RBLOCK_SIZE_STREAMED;
	iov[1].iov_base	= payloadbuf;
	iov[1].iov_len	= bufsize;
	msg_hdr.msg_iov	= iov;

	if (bufsize > 0) {
		assert(payloadbuf != NULL);
		msg_hdr.msg_iovlen = 2;
		block->payload = payloadbuf;
	} else {
		msg_hdr.msg_iovlen = 1;
		block->payload = NULL;
	}

	size = recvmsg(fd, &msg_hdr, flags);
	if (size <= 0) {
		/* error or shutdown */
		return size;
	}

	/* now, for some more sanity checks: */
	if (size < RBLOCK_SIZE_STREAMED) {
		/* block header not fully queued - error out */
		printf("error: block header not fully queued.\n");
		return -1;
	}
	if (size < RBLOCK_SIZE_STREAMED + block->payload_size) {
		/* too little bytes queued. error out */
		printf("warning: block not fully queued.\n");
		return -1;
	}
	if (msg_hdr.msg_flags & MSG_TRUNC) {
		printf("warning: buffer given to rblock_recv_into() too small !\n");
		return -1;
	}

	return 1;
}
#endif

t_rblock* rblock_recv(int fd, int flags) {
	t_rblock* block;
	int err;

	block = rblock_new(0);
	err = rblock_recv_into(fd, block, flags, NULL, 0);
	if (err<=0) {
		rblock_free(block);
		return NULL;
	}
	return block;
}

void _dump_format(FORMAT *avformat) {
  	printf("FORMAT.width: %d\n",avformat->width);
  	printf("FORMAT.height: %d\n",avformat->height);
  	printf("FORMAT.is_bw: %d\n",avformat->is_bw);
  	printf("FORMAT.block_anz: %d\n",avformat->block_anz);
  	printf("FORMAT.bytes_per_block: 0x%x (%d)\n",
  		   avformat->bytes_per_block,
  		   avformat->bytes_per_block);
  	printf("FORMAT.audio_buffer_size: 0x%x (%d)\n",
  		   avformat->audio_buffer_size,
  		   avformat->audio_buffer_size);
}

void rblock_dump(t_rblock *block) {
	if (block == NULL) {
		printf("error: rblock_dump() called with NULL block !");
		return;
	}
	printf("block: fno: %04d, size %4d (0x%4x), type ", 
			block->filenr,
			block->payload_size, 
			block->payload_size);
	switch (block->header.marker) {
		case BLOCK_VIDEO:
			printf("VIDEO, block_num: %4d",
				block->header.video.block_num);
			if (block->payload_size != block->header.video.block_size) {
				printf(", block_size: 0x%4x !!!",
					block->header.video.block_size);
			}
			printf("\n");
			break;
		case BLOCK_AUDIO:
			printf("AUDIO\n");
			break;
		case BLOCK_FORMAT:
			printf("FORMAT\n");
			_dump_format(&(block->header.format.avformat));
			break;
		case BLOCK_AUDIO_ACK:
			printf("AUDIO_ACK, send_up_to: %d, ack:%d\n",
					block->header.status.send_up_to,
					block->header.status.acked_file);
			break;
		case BLOCK_REQ:
			printf("REQ, send_up_to: %d\n",
					block->header.status.send_up_to);
			break;
		case BLOCK_HELLO:
			printf("HELLO\n");
			break;
		default:
			printf("OTHER (%d) !!!\n", block->header.marker);
	}
}

void rblock_dump_payload(t_rblock* block) {
	char *payload	= block->payload;
	int size		= block->payload_size;
	int i;
	printf("------------- payload --------------\n");
	rblock_dump(block);
	for(i=0; i<size; ++i){
  		if(i%16==0) printf("\n");
		if(i%2==0) printf(" ");
		printf("%02x ",(unsigned char)payload[i]);
	}
	printf("\n");
}

/******************************************************************************/
/* alternate implementaions as workaround for buggy Libnetbug: :-)            */
/******************************************************************************/

#ifdef WORKAROUND_BUGGY_LIBNETBUG
size_t rblock_send(int fd, t_rblock* block, int flags) {
	char buf[4096];

	if (block == NULL) {
		return -1;
	}

	memcpy(buf, block, RBLOCK_SIZE_STREAMED);
	if ( (block->payload==NULL) || (block->payload_size == 0)) {
	} else {
		memcpy(buf+RBLOCK_SIZE_STREAMED, block->payload, block->payload_size);
	}

	return send(fd, buf, RBLOCK_SIZE_STREAMED + block->payload_size, flags);
}

int rblock_recv_into(int fd, t_rblock* block, int flags, char payloadbuf[], size_t bufsize) {
	size_t size;
	char buf[4096];

	size = recv(fd, buf, 4096, flags);
	if (size <= 0) {
		/* error or shutdown */
		return size;
	}

	/* now, for some more sanity checks: */
	if (size < RBLOCK_SIZE_STREAMED) {
		/* block header not fully queued - error out */
		printf("error: block header not fully queued.\n");
		return -1;
	}

	memcpy(block, buf, RBLOCK_SIZE_STREAMED);
	if ( block->payload_size != 0) {
		assert(payloadbuf != NULL);
		memcpy(payloadbuf, buf+RBLOCK_SIZE_STREAMED, block->payload_size);
		block->payload = payloadbuf;
	} else {
		block->payload = NULL;
	}

	if (size < RBLOCK_SIZE_STREAMED + block->payload_size) {
		/* too little bytes queued. error out */
		printf("warning: block not fully queued.\n");
		return -1;
	}
	return 1;
}
#endif
/* vi:set tabstop=4 shiftwidth=4 noexpandtab: */
