/* Framework for the Audio/Video Sync Task 
 * Praktikum Kommunikationssysteme 2004
 *
 * Lars Martensen <l.martensen@gmx.de>
 */
#include "video.h"

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <time.h>

#include <jpeglib.h>
#include <jerror.h>
#include <setjmp.h>

#define TABLE_ROWS 5
#define TABLE_COLUMNS 3
#define WINDOW_WIDTH 400
#define WINDOW_HEIGHT 350

#ifdef __sparc__

#ifdef __arch64__
#include <asm-sparc64/audioio.h>
#else
#include <asm-sparc/audioio.h>
#endif

#define MAX_VOLUME AUDIO_MAX_GAIN
static int audio_mode = AUDIO_SPEAKER;

#else

#define MONO 1
#define SAMPLES_PER_SECOND 8000
#define MAX_VOLUME 100.0
static int mixerfd;
#endif

/* main_loop will be called every TIMEOUT miliseconds */
#define TIMEOUT 2

/* we use 24 Bit RGB */
#define BYTES_PER_PIXEL 3
#define MAX_PICNUM 80

extern void init(char *hostName, int portNr, char *movie);
extern void main_loop();

extern void ptm_buffer_src(j_decompress_ptr cinfo);
extern void ptm_buffer_set(j_decompress_ptr cinfo, char* buf, int len);

static int audiofd;
static FORMAT format;
static GtkWidget *gtktable = NULL;
static GdkPixmap *gdkpixmap = NULL;
static GtkPixmap *gtkpixmap = NULL;
static GdkGC *gdkgc = NULL;

static char *image_data = NULL;

static struct jpeg_decompress_struct cinfo;
static struct jpeg_error_mgr jerr;

static void die(const char *s, ...)
{
	va_list ap;
	va_start(ap, s);
	
	if (s) { 
		vfprintf(stderr, s, ap);
		fprintf(stderr, "\n");
	}
	va_end(ap);

	exit(EXIT_FAILURE);
}

static inline void *xmalloc(size_t size)
{
	void *p; 

	p = malloc(size);
	if (!p)
		die("malloc: out of memory");
	
	return p;
}
	
static gboolean delete_event(GtkWidget *widget, GdkEvent *event, 
			     gpointer data)
{
	gtk_main_quit();
	return FALSE;
}

static void destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

void graceful_exit()
{
	gtk_main_quit();
}




static void init_audio()
{
#ifdef __sparc__
	audio_info_t audioInfo;
	audiofd = open("/dev/audio", O_WRONLY);
	if (audiofd < 0) {
		perror("open audio");
		die(NULL);
	}
  	/* Daten in der Struktur werden auf Werte gesetzt, die vom Audiodevice */
  	/* ignoriert werden                                                    */
  	AUDIO_INITINFO(&audioInfo);
    	/* Kopfhoerer als Standartausgabe einstellen */
    	audioInfo.play.port = AUDIO_HEADPHONE; 

    	/* Lautstaerke einstellen */
    	audioInfo.play.gain =  100;
    	/* Daten im Audiodevice einstellen */
    	ioctl(audiofd, AUDIO_SETINFO, &audioInfo);

#else	
	int format = AFMT_MU_LAW;
	int caps;
	
	audiofd = open("/dev/dsp", O_WRONLY);
	if (audiofd < 0) {
		perror("open audio");
		die(NULL);
	}

	ioctl(audiofd, SNDCTL_DSP_GETCAPS, &caps);
	if ( !(caps & DSP_CAP_REALTIME))
		fprintf(stderr, "Warning: DSP_CAP_REALTIME not supported!\n");
	if (caps & DSP_CAP_BATCH)
		fprintf(stderr, "Warning: DSP_CAP_BATCH supported! "
		       "Timing may be inaccurate\n");

	mixerfd = open("/dev/mixer", O_RDWR);
	if (mixerfd < 0) {
		perror("open mixer");
		die(NULL);
	}

	if (ioctl(audiofd, SNDCTL_DSP_SETFMT, &format) == -1) {
		perror("SNDCTL_DSP_SETFMT");
		die(NULL);
	}

	if (format != AFMT_MU_LAW) {
		fprintf(stderr, "MU_LAW not supported!\n");
		die(NULL);
	}

	format = MONO;
	if (ioctl(audiofd, SNDCTL_DSP_CHANNELS, &format) == -1) {
		perror("SNDCTL_DSP_CHANNELS");
		die(NULL);
	}

	format = SAMPLES_PER_SECOND;
	if (ioctl(audiofd, SNDCTL_DSP_SPEED, &format) == -1) {
		perror("SNDCTL_DSP_SPEED");
		die(NULL);
	}
#endif
}

static int get_volume()
{
#ifdef __sparc__
	audio_info_t info;

	ioctl(audiofd, AUDIO_GETINFO, &info);

	return (int)info.play.gain;
#else	
	int vol;

	ioctl(mixerfd, SOUND_MIXER_READ_PCM, &vol);
	vol &= 0x00ff;

	return vol;
#endif
}

static void set_volume(int vol)
{
#ifdef __sparc__
	audio_info_t info;
	
	AUDIO_INITINFO(&info);
	info.play.port = audio_mode;
	info.play.gain = vol;

	ioctl(audiofd, AUDIO_SETINFO, &info);
#else
	/* set volume for right channel too */
	vol = ((vol & 0xff) << 8) | vol;

	ioctl(mixerfd, SOUND_MIXER_WRITE_PCM, &vol);
#endif
}

/* gets called when using the volume hscale */
static void volume_changed(GtkRange *range, gpointer user_data)
{
	int vol;
	GtkToggleButton *mute_button = GTK_TOGGLE_BUTTON(user_data);
	
	/* do nothing when muted */
	if (gtk_toggle_button_get_active(mute_button))
		return;
	
	vol = (int)gtk_range_get_value(range);
	set_volume(vol);
}

/* to play exactly one audio block */
int play_audio_block(char *data, size_t length)
{
	int res;

	res = write(audiofd, data, length);
#ifdef __sparc__
	/* write a zero length buffer as described in audio(7I) */
	write(audiofd, data, 0);
#endif 
	return res;
}

/* returns the number of already played audio blocks */
int get_completed_audio_blocks()
{
#ifdef __sparc__
	audio_info_t info;

	if (ioctl(audiofd, AUDIO_GETINFO, &info) == -1) {
		perror("ioctl AUDIO_GETINFO");
		die(NULL);
	}

	return (int)info.play.eof;
#else
	count_info info;

	if (ioctl(audiofd, SNDCTL_DSP_GETOPTR, &info) == -1) {
		perror("ioctl SNDCTL_DSP_GETOPTR");
		die(NULL);
	}

	return (info.bytes / format.audio_buffer_size);
#endif
}


void init_audio_video(FORMAT *video, int picnum)
{
	int depth;

	if (picnum > MAX_PICNUM) {
		die("init_video: picnum must not be greater than %d", 
		    MAX_PICNUM);
	}

	memcpy(&format, video, sizeof(FORMAT));

	/* create low level drawable */
	depth = gdk_visual_get_best_depth();
	gdkpixmap = gdk_pixmap_new(NULL, format.width, format.height, depth);

	/* create high level drawable */
	gtkpixmap = GTK_PIXMAP(gtk_pixmap_new(gdkpixmap, NULL));
	gtk_table_attach_defaults(GTK_TABLE(gtktable), GTK_WIDGET(gtkpixmap),
				  0, TABLE_COLUMNS, 0, TABLE_ROWS-1);
	gtk_widget_show(GTK_WIDGET(gtkpixmap));

	/* needed by gdk_draw_rgb_image() in show_video() */
	gdkgc = gdk_gc_new(gdkpixmap);
	
	/* here the decompressed jpegs will be stored */
	image_data = xmalloc(format.width * format.height * 
			    BYTES_PER_PIXEL * picnum);
	
	/* some libjpeg init stuff */
	memset(&cinfo, 0, sizeof(cinfo));
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	ptm_buffer_src(&cinfo);
}

/* draws one picture */
void show_video(int offset)
{
	/* if init_video was not yet called */
	if (!gtkpixmap)
		return;

	/* draw on gdkpixmap */
	gdk_draw_rgb_image(gdkpixmap, gdkgc, 0, 0, format.width, format.height, 
			   GDK_RGB_DITHER_NONE, 
			   image_data + format.width * BYTES_PER_PIXEL * offset, 
			   format.width * BYTES_PER_PIXEL);
	/* redraw gtkpixmap */
	gtk_widget_queue_draw(GTK_WIDGET(gtkpixmap));
}

/* this is the most timing critical function, 
 * modify only if you know what you're doing
 */
int decomp_JPEG (char* data, int data_size, int offset)
{
	static char **parray = NULL;
	int i, width, height, lines, restLines;
	char *bufp, **p;

	/* this tells libjpeg how to get the jpeg data */
	ptm_buffer_set(&cinfo, data, data_size);

	jpeg_read_header(&cinfo, TRUE);

	/* make decompression fast */	
	cinfo.two_pass_quantize = FALSE;
	cinfo.dither_mode = JDITHER_ORDERED;
	cinfo.dct_method = JDCT_FASTEST;
	cinfo.do_fancy_upsampling = FALSE;
	cinfo.do_block_smoothing = FALSE;

	jpeg_start_decompress(&cinfo);
	
	width = cinfo.output_width * cinfo.output_components;
	height = cinfo.output_height;

	/* create an array for the scanlines */
	if (!parray)
		parray = xmalloc(height * sizeof(char*));

	/* initialize scanlines array */
	bufp = image_data + offset * BYTES_PER_PIXEL;
	for (i = 0; i < height; i++) {
		parray[i] = bufp;
		bufp += width;
	}

	/* let libjpeg do the work */
	p = parray;
	restLines = height;
	while (cinfo.output_scanline < cinfo.output_height) {
		lines = jpeg_read_scanlines(&cinfo, (JSAMPARRAY)p, restLines);
		p += lines;
		restLines -= lines;
	}

	jpeg_finish_decompress(&cinfo);
	
	return 1; /* success */
}

static gint main_loop_wrapper(gpointer data)
{
	gdk_threads_enter();
	main_loop();
	gdk_threads_leave();
	return TRUE;
}

static void mute_button_toggled(GtkToggleButton *button, gpointer *user_data)
{
	if (gtk_toggle_button_get_active(button))
		set_volume(0);
}

static void headphone_button_toggled(GtkToggleButton *button, 
				    gpointer *user_data)
{
#ifdef __sparc__
	if (gtk_toggle_button_get_active(button)) {
		audio_mode = AUDIO_HEADPHONE;
		set_volume(get_volume());
	}
#endif
}

static void speaker_button_toggled(GtkToggleButton *button, 
				    gpointer *user_data)
{
	int vol;
	GtkRange *volume = GTK_RANGE(user_data);
	
	vol = (int)gtk_range_get_value(volume);
	
	if (gtk_toggle_button_get_active(button)) {
#ifdef __sparc__
		audio_mode = AUDIO_SPEAKER;
		set_volume(get_volume());
#else
		set_volume(vol);
#endif
	}
}

static GtkWidget *create_window(int width, int height)
{
	GtkWidget *gtkwindow;
	
	gtkwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(GTK_WINDOW(gtkwindow), width, 
				    height);

	g_signal_connect(G_OBJECT(gtkwindow), "delete_event",
			 G_CALLBACK(delete_event), NULL);
	g_signal_connect(G_OBJECT(gtkwindow), "destroy", 
			 G_CALLBACK(destroy), NULL);

	gtk_container_set_border_width(GTK_CONTAINER(gtkwindow), 10);

	return gtkwindow;
}

int main(int argc, char **argv)
{
	GtkWidget *gtkwindow;
	GtkWidget *quit_button, *mute_button, 
		  *headphone_button, *speaker_button;
	GtkWidget *volume;
	GtkWidget *vbox1, *vbox2;
	GtkWidget *label;
	int vol;

	/* gtk init stuff */
	g_thread_init(NULL);
	gdk_threads_init();
	gtk_init(&argc, &argv);
	
	/* open audio/mixer devices */
	init_audio();
	vol = get_volume();

	if (argc != 4) {
		die("Usage: %s <hostname> <port> <movie>", argv[0]);
	}
	init(argv[1], atoi(argv[2]), argv[3]);
	
	/* create main window */
	gtkwindow = create_window(WINDOW_WIDTH, WINDOW_HEIGHT);

	/* all widgets will be inserted into this table */
	gtktable = gtk_table_new(TABLE_ROWS, TABLE_COLUMNS, FALSE);
	gtk_container_add(GTK_CONTAINER(gtkwindow), gtktable);
	gtk_widget_show(gtktable);

	/* quit button */
	quit_button = gtk_button_new_with_label("Quit");
	g_signal_connect(G_OBJECT(quit_button), "clicked",
			 G_CALLBACK(destroy), NULL);
	gtk_table_attach_defaults(GTK_TABLE(gtktable), quit_button, 0, 1, 
				  TABLE_ROWS-1, TABLE_ROWS);
	gtk_widget_show(quit_button);
	
	/* vbox for "Volume" label and volume hscale */
	vbox2 = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox2);

	/* "Volume" label */
	label = gtk_label_new("Volume");
	gtk_box_pack_start(GTK_BOX(vbox2), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	/* volume hscale */
	volume = gtk_hscale_new_with_range(0.0, MAX_VOLUME, 1.0);
	gtk_range_set_value(GTK_RANGE(volume), (double)vol);
	gtk_box_pack_start(GTK_BOX(vbox2), volume, TRUE, TRUE, 0);
	gtk_widget_show(volume);
	
	/* vbox for volume buttons */
	vbox1 = gtk_vbox_new(TRUE, 0);
	gtk_widget_show(vbox1);
	
	/* mute button */
	mute_button = gtk_radio_button_new_with_label(NULL, "Mute");
	gtk_box_pack_start(GTK_BOX(vbox1), mute_button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(mute_button), "toggled", 
			 G_CALLBACK(mute_button_toggled), NULL);
	gtk_widget_show(mute_button);
	
	/* headphone button */
	headphone_button = gtk_radio_button_new_with_label_from_widget(
			GTK_RADIO_BUTTON(mute_button), "Headphone");
	gtk_box_pack_start(GTK_BOX(vbox1), headphone_button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(headphone_button), "toggled", 
			 G_CALLBACK(headphone_button_toggled), NULL);
	gtk_widget_show(headphone_button);
	
	/* speaker button */
	speaker_button = gtk_radio_button_new_with_label_from_widget(
			GTK_RADIO_BUTTON(headphone_button), "Speaker");
	gtk_box_pack_start(GTK_BOX(vbox1), speaker_button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(speaker_button), "toggled", 
			 G_CALLBACK(speaker_button_toggled), volume);
	gtk_widget_show(speaker_button);

	/* default is headphone */ //ZK
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(headphone_button), TRUE);
	
	g_signal_connect(G_OBJECT(volume), "value-changed", 
			 G_CALLBACK(volume_changed), mute_button);


	gtk_table_attach_defaults(GTK_TABLE(gtktable), vbox1, 1, 2, 
				  TABLE_ROWS-1, TABLE_ROWS);
	gtk_table_attach_defaults(GTK_TABLE(gtktable), vbox2, 2, 3, 
				  TABLE_ROWS-1, TABLE_ROWS);

	/* show the main window */	
	gtk_widget_show(gtkwindow);

	/* start calling main_loop every TIMEOUT miliseconds */
	g_timeout_add(TIMEOUT, main_loop_wrapper, NULL); 

	gtk_main();

	return 0;
}
